from requests_oauthlib import OAuth2Session
from oauthlib.oauth2 import BackendApplicationClient
from collections.abc import Callable
from typing import Optional
from requests import Response
from utils import slugify
# import jose

from .oauth import auth_host

try:
    import psycopg
    import psycopg_binary
except ImportError:
    try:
        import mysql.connector
    except ImportError:
        try:
            import pymongo
        except ImportError:
            try:
                import redis
            except ImportError:
                try:
                    import sqlite3
                except ImportError:
                    Cache = None
                else:
                    from battlenet_client.cache.sqlite import SQLiteCache as Cache
            else:
                from battlenet_client.cache.redis import RedisCache as Cache
        else:
            from battlenet_client.cache.mongodb import MongoDBCache as Cache
    else:
        from battlenet_client.cache.mysql import MySQLCache as Cache
else:
    from battlenet_client.cache.postgresql import PostgresCache as Cache


class BattlenetClient:
    oauth = None
    state = None
    region = None
    client_id = None
    client_secret = None
    redirect_uri = None
    cache = None

    def __init__(self, region: str, oauth: OAuth2Session, cache_uri: Optional[str] = None):
        self.region = region
        self.oauth = oauth

        if cache_uri:
            self.cache = Cache(cache_uri)

    @classmethod
    def client_credential(cls, region: str,  client_id: str, client_secret: str, *,
                          auto_refresh_url: Optional[str] = None, auto_refresh_kwargs: Optional[dict] = None,
                          cache_uri: Optional[str] = None):
        """ Creates an instance of client credential grant oauth2 session

        Args:
            region (str): The region where the client is connected
            client_id (str): Client ID issued by develop.battle.net
            client_secret (str): Client secret issued by develop.battle.net
            auto_refresh_url (Optional[str]): Auto-refresh URL when token expires
            auto_refresh_kwargs (Optional[dict]): Extra arguments passed to OAuth2Session for automatic token refresh
            cache_uri (Optional[str]): URI for the cache system

        Returns:
            instance of BattlenetClient configured for client credentials grant oauth2 session
        """
        cls.client_id = client_id
        cls.client_secret = client_secret


        if auto_refresh_url and auto_refresh_kwargs:
            oauth = OAuth2Session(client=BackendApplicationClient(client_id=client_id),
                                  auto_refresh_url=auto_refresh_url,
                                  auto_refresh_kwargs=auto_refresh_kwargs,
                                  update_token=cls.token_updater)
        else:
            oauth = OAuth2Session(client=BackendApplicationClient(client_id=client_id))

        oauth.fetch_token(token_url=f"{auth_host(region)}/token", client_id=client_id, client_secret=client_secret)
        return cls(region, oauth, cache_uri)

    @classmethod
    def authorization_code(cls, region, client_id: str, redirect_uri: str, scope: list[str], *,
                           auto_refresh_url: Optional[str] = None, auto_refresh_kwargs: Optional[dict] = None,
                           updater: Callable[[dict], None] = None, cache_uri: Optional[str] = None):
        """ Creates an instance of authorization code grant oauth2 session

        Args:
            region (str): The region where the client is connected
            client_id (str): Client ID issued by develop.battle.net
            redirect_uri (str): URL to send to authorization service to return
            scope (list[str]): list of scopes to grant
            auto_refresh_url (Optional[str]): Auto-refresh URL when token expires
            auto_refresh_kwargs (Optional[dict]): Extra arguments passed to OAuth2Session for automatic token refresh
            updater (callable, optional): the function to perform the automatic update
            cache_uri (Optional[str]): URI for the cache system

        Returns:
            instance of BattlenetClient configured for authorization code grant oauth2 session
        """

        oauth = OAuth2Session(client_id, redirect_uri=redirect_uri, scope=scope, auto_refresh_url=auto_refresh_url,
                              auto_refresh_kwargs=auto_refresh_kwargs, updater=updater)



        return cls(region, oauth, cache_uri)

    @classmethod
    def open_id(cls):
        """TODO: Future release"""
        pass

    def get_authorization_url(self) -> tuple[str, str]:
        """ Creates an authorization url grant oauth2 URL

        Returns:
              tuple[str, str]: authorization url and authorization state
        """
        url, self.state = self.oauth.authorization_url(f"{auth_host(self.region)}/authorize")
        return url, self.state

    def callback(self, auth_uri: str, client_secret: str, response: Response) -> None:
        """ Creates an authorization code grant oauth2 session

        Args:
            auth_uri (str): URL to acquire the authorization url
            client_secret (str): Client secret issued by develop.battle.net
            response (Response): response from authorization server
        """
        self.oauth.fetch_token(auth_uri, authorization_response=response, client_secret=client_secret)

    def token_updater(self, token: dict):
        """ Function to update the token after a refresh

        Args:
            token (dict): token returned from authorization server
        """
        self.oauth.token = token
        self.oauth.access_token = token["access_token"]

    def fetch_token(self, scope: str, grant_type: str = 'client_credentials',
                    redirect_uri: Optional[str]=None) -> None:
        """ Fetches the token from the authorization server

        Args:
            scope (str): scope for the authorization
            grant_type (str, optional): grant type for the authorization defaults to 'client_credentials'
            redirect_uri (str, optional): redirect uri for the authorization server
        """
        data = {"region": self.region.lower(), "grant_type": grant_type, "scope": scope, "code": "code", "redirect_uri": redirect_uri}

        self.oauth.post(f"{auth_host(self.region)}/token", data=data)

    def user_info(self):
        """Returns basic information about the user associated with the current bearer token."""
        return self.oauth.get(f"{auth_host(self.region)}/userinfo")

    def check_token(self) -> bool:
        """ Verifies that a given bearer token is valid and retrieves metadata about the token, including the client_id
        used to create the token, expiration timestamp, and scopes granted to the token.

        Returns:
            bool: True if the token is valid, False otherwise
        """
        return self.oauth.post(f"{auth_host(self.region)}/check_token",
                               data={"token": self.oauth.token['access_token']}).status_code == 200

    def get(self, api_endpoint: Callable, category_name: str,  *args, **kwargs) -> Optional[bytes]:
        """ Retrieves data from the API endpoint or cache

        Args:
            api_endpoint (function): API endpoint function from the respective submodule
            category_name (str): category name
            *args: list of positional arguments for the api endpoint function
            **kwargs: dict of keyword arguments for the api endpoint function

        Returns:
            bytes: data returned from the API endpoint
        """
        if self.cache.check(self.region, slugify(category_name), *args, **kwargs):
            return self.cache.select(self.region, slugify(category_name), *args, **kwargs)

        url, params = api_endpoint(self.region, *args, **kwargs)

        response = self.oauth.get(url, params=params)

        if self.cache.upsert(self.region, slugify(category_name), *args, data=response.content, **kwargs) > 0:
            return response.content

        return None

    def post(self, url, **kwargs):
        """Conveneince function for POST requests"""
        return self.oauth.post(url, **kwargs)

    def put(self, url, **kwargs):
        """Conveneince function for PUT requests"""
        return self.oauth.put(url, **kwargs)

    def delete(self, url, **kwargs):
        """Conveneince function for DELETE requests"""
        return self.oauth.delete(url, **kwargs)

    def add_cache(self, cache_uri: str):
        """ Initializes connection for cache

        Args:
            cache_uri (str): URI for the cache system
        """
        self.cache = Cache(cache_uri)



import pytest

from src.battlenet_client.constants import VALID_REGIONS
from src.battlenet_client.oauth import user_info, token_validation
from src.battlenet_client.exceptions import BNetRegionNotFoundError

from tests.constants import INVALID_REGIONS


@pytest.mark.parametrize('region_tag', VALID_REGIONS)
def test_user_info(region_tag):
    data = user_info(region_tag, locale='enus')
    assert isinstance(data, tuple)
    assert isinstance(data[0], str)
    if region_tag == 'cn':
        assert data[0] == "https://www.battlenet.com.cn/oauth/userinfo"
    elif region_tag in ('tw', 'kr'):
        assert data[0] == f"https://apac.battle.net/oauth/userinfo"
    else:
        assert data[0] == f'https://{region_tag.lower()}.battle.net/oauth/userinfo'
    assert 'locale' in data[1]
    assert data[1]['locale'] == 'en_US'


@pytest.mark.parametrize('region_tag', INVALID_REGIONS)
def test_user_info_invalid_region(region_tag):
    with pytest.raises(BNetRegionNotFoundError):
        user_info(region_tag, locale='enus')


@pytest.mark.parametrize('region_tag', VALID_REGIONS)
def test_token_validation(region_tag):
    token = 'my_good_token_1234'
    data = token_validation(region_tag, token, locale='enus')
    assert isinstance(data, tuple)
    assert isinstance(data[0], str)
    if region_tag == 'cn':
        assert data[0] == "https://www.battlenet.com.cn/oauth/check_token"
    elif region_tag in ('tw', 'kr'):
        assert data[0] == f"https://apac.battle.net/oauth/check_token"
    else:
        assert data[0] == f'https://{region_tag.lower()}.battle.net/oauth/check_token'
    assert isinstance(data[1], dict)
    assert 'locale' in data[1]
    assert 'token' in data[1]
    assert data[1]['locale'] == 'en_US'


@pytest.mark.parametrize('region_tag', INVALID_REGIONS)
def test_token_validation_invalid_region(region_tag):
    with pytest.raises(BNetRegionNotFoundError):
        token = 'my_good_token_1234'
        token_validation(region_tag, token, locale='enus')



import sqlite3
import os

from typing import Union, Optional

from .base import BaseCache


class SQLiteCache(BaseCache):

    connector = None
    cursor = None

    def __init__(self, db_path = "file:memory?mode=memory") -> None:
        self.connector = sqlite3.connect(db_path, uri=True)

    def __del__(self) -> None:
        self.connector.close()

    def check(self, region_tag: str, category_tag: str, object_id: Union[str, int], locale: str, release: Optional[str]='retail') -> bool:
        """ Used to check if a valid cache entry exists.

        Args:
            region_tag (str): region tag for the API
            category_tag (str): category tag for the API
            object_id (Union[str, int]): object ID of the requested item
            locale (str): locale for the request
            release (str, optional): release type for the game
        Returns:
            bool: True if the cache entry exists, False if doesn't exist or expired
        """

        hash_value = self.get_hash(region_tag, category_tag, release, object_id, locale)
        try:
            res = self.connector.execute("""SELECT COUNT(*) as count
FROM cache, categories
WHERE cache.object_hash = ? and categories.name = ? and cache.category = categories.cat_id and 
unixepoch('now') - cache.last_updated <= categories.duration;""",
                                   (hash_value, category_tag)).fetchone()
        except TypeError:
            return False
        else:
            return res[0] > 0

    def select(self, region_tag: str, category_tag: str, object_id: Union[str, int], locale: str,
               release: Optional[str]='retail') -> Union[bytes|None]:
        """ Perform cached data from the selected record(s)

        Args:
            region_tag (str): region tag for the API
            category_tag (str): category tag for the API
            object_id (Union[str, int]): object ID of the requested item
            locale (str): locale for the request
            release (str, optional): release type for the game

        Returns:
            bytes or None: content of the requested cache if it exists or None otherwise
        """

        hash_value = BaseCache.get_hash(region_tag, category_tag, release, object_id, locale)

        if not self.check(region_tag, category_tag, object_id, locale):
            return None

        try:
            res = self.connector.execute("""SELECT data 
FROM cache, categories
WHERE cache.object_hash = ? and categories.name = ? and cache.category = categories.cat_id and 
unixepoch('now') - cache.last_updated <= categories.duration order by seq;""",
                                      (hash_value, category_tag)).fetchall()
        except TypeError:
            return None
        else:
            if len(res) == 0:
                return None

            return b''.join([row[0] for row in res])

    def upsert(self, region_tag: str, category_tag: str, object_id: Union[str, int], locale: str,
               data: bytes, chunk_size: int=32 * 1024, release: Optional[str]='retail') -> int:
        """ Insert or update the data within the given record

        Args:
            region_tag (str): region tag for the API
            category_tag (str): category tag for the API
            object_id (Union[str, int]): object ID of the requested item
            locale (str): locale for the request
            data (bytes): data to be inserted into the cache
            chunk_size (int): size of the chunk to insert into the cache
            release (str, optional): release type for the game

        Returns:
            int: returns the number of inserted/updated rows
        """

        hash_value = BaseCache.get_hash(region_tag, category_tag, release, object_id, locale)
        count = 1
        data_list = []
        for chunk in BaseCache.chunk_data(data, chunk_size):
            data_list.append({'hash': hash_value, "sequence": count, "data": chunk, "name": category_tag})
            count += 1

        try:
            res = self.connector.executemany("""INSERT INTO cache (object_hash, seq, data, category, last_updated)
SELECT :hash, :sequence, :data, cat_id, unixepoch('now') FROM categories WHERE name=:name
ON CONFLICT (object_hash, seq) DO UPDATE SET data = excluded.data, category=excluded.category, last_updated = excluded.last_updated;""",
                            data_list)
        except TypeError:
            return False
        else:
            self.connector.commit()
            return res.rowcount


    def delete(self, region_tag: str, category_tag: str, object_id: Union[str, int], locale: str,
               release: Optional[str]='retail') -> int:
        """ Delete the records that match given criteria

        Args:
            region_tag (str): region tag for the API
            category_tag (str): category tag for the API
            object_id (Union[str, int]): object ID of the requested item
            locale (str): locale for the request
            release (str, optional): release type for the game

        Returns:
            int: returns the number of deleted rows
        """

        delete_dict = {"hash": self.get_hash(region_tag, category_tag, release, object_id, locale)}

        try:
            res = self.connector.execute("""DELETE FROM cache WHERE object_hash = :hash;""",
                delete_dict)
        except TypeError:
            return False
        else:
            self.connector.commit()
            return res.rowcount



from abc import ABC, abstractmethod
from hashlib import sha1
from typing import Union, Optional


class BaseCache(ABC):

    @staticmethod
    def get_hash(region_tag: str, catagory_tag: str, release:str , object_id: Union[str,int], locale: Optional[str]) -> str:

        if not isinstance(region_tag, str):
            raise TypeError("region tag must be a string")

        if not isinstance(catagory_tag, str):
            raise TypeError("catagory tag must be a string")

        if not isinstance(object_id, (str, int)):
            raise TypeError("object id must be a string or int")

        if not isinstance(release, str):
            raise TypeError("release must be a string")

        return sha1(f"{region_tag}|{catagory_tag}|{release}|{object_id}|{locale}".encode()).hexdigest()

    @staticmethod
    def chunk_data(data: bytes, chunk_size: int=60*1024) -> list[bytes]:

        if not isinstance(data, bytes):
            raise TypeError("Data must be of type bytes")
        if not isinstance(chunk_size, int):
            raise TypeError("Chunk size must be integer")
        if chunk_size < 1 or chunk_size > 61_440:
            raise ValueError("Chunk size must be between 1 and 61_440")
        if len(data) < 1:
            raise ValueError("data size must be at least 1 byte")

        chunks = []
        for chunk in range(0, len(data), chunk_size):
            chunks.append(data[chunk:chunk+chunk_size])
        return chunks

    @abstractmethod
    def check(self, region_tag: str, catagory_tag: str, release: str, object_id: str, locale: str) -> bool:
        raise NotImplementedError("This method is not implemented")

    @abstractmethod
    def select(self, region_tag: str, catagory_tag: str, release: str, object_id: str, locale: str) -> bytes:
        raise NotImplementedError("This method is not implemented")

    @abstractmethod
    def upsert(self, region_tag: str, catagory_tag: str, release: str, object_id: str, locale: str, data: bytes) -> int:
        raise NotImplementedError("This method is not implemented")

    @abstractmethod
    def delete(self, region_tag: str, catagory_tag: str, release: str, object_id: str, locale: str) -> int:
        raise NotImplementedError("This method is not implemented")



from .base import BaseCache

class MongoDBCache(BaseCache):
    def check(self, region_tag: str, catagory_tag: str, release: str, object_id: str, locale: str) -> bool:
        pass

    def select(self, region_tag: str, catagory_tag: str, release: str, object_id: str, locale: str) -> list:
        pass

    def upsert(self, region_tag: str, catagory_tag: str, release: str, object_id: str, locale: str, data: bytes) -> int:
        pass

    def delete(self, region_tag: str, catagory_tag: str, release: str, object_id: str, locale: str) -> int:
        pass



"""Defines the functions that generate the URLs, parameter lists, and headers
for the OAuth API Endpoints

Functions:
    user_info(region_tag, locale)
    validate_token(region_tag, token, locale)

Misc Variables:
    __version__
    __author__

Author: David "Gahd" Couples
License: GPL v3
Copyright: February 24, 2022
"""

from .utils import *
from .decorators import verify_region


__version__ = "1.0.3"
__author__ = "David \"Gahd\" Couples"


@verify_region
def authorization(region_tag: str, client_id: str, redirect_uri: str, scope: str,
                  state: str, response_type: Optional[str]="code") -> tuple[str, dict]:
    """Required for authorization code flow

    Args:
        state (str): value suppplied to maintain state between request and callback
        scope (str): list of scopes to authorize
        redirect_uri (str): callback url
        client_id (str): client id provided by API access
        region_tag (str): region of the data request
        response_type (str, optional): response type (typically `code`)

    Returns:
        tuple(str, dict): returns the tuple of url, data, and params
    """
    url = f"{auth_host(region_tag)}/authorize"

    data = {"response_type": response_type, "client_id": client_id, "redirect_uri": redirect_uri, "scope": scope,
            "state": state}

    return url, data


@verify_region
def fetch_token(region_tag: str, grant_type: str, client_id:str, *, code: Optional[str]=None,
          redirect_uri: Optional[str]=None) -> tuple[str, dict]:
    """ Retrieves the token from a authorization code flow

    Args:
        region_tag (str): region of the data request
        grant_type (str): grant type , typically `authorization_code`
        code (str): authorization code received from the authorization server
        redirect_uri (str): return URL, needs to match the original request
        client_id (str): client id provided by API access

    Returns:
        tuple(str, dict, dict): the url, params, and data for retrieving the token
    """
    url = f"{auth_host(region_tag)}/token"

    data = {"grant_type": grant_type}

    if grant_type == "authorization_code":
        data["code"] = code
        data["client_id"] = client_id
        data["redirect_uri"] = redirect_uri

    return url, data


@verify_region
def user_info(region_tag: str, locale: Optional[str] = None) -> tuple:
    """Returns the user info

    Args:
        region_tag (str): region_tag abbreviation
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    url = f"{auth_host(region_tag)}/userinfo"

    params = {"locale": localize(locale)}

    return url, params


@verify_region
def token_validation(region_tag: str,  token: str, locale: Optional[str] = None) -> tuple[str, dict, dict]:
    """Returns if the token is still valid or not

    Args:
        region_tag (str): region_tag abbreviation
        token (str): token string to validate
        locale (str): which locale to use for the request

    Returns:
        tuple: The URL (str) and parameters (dict)
    """
    url = f"{auth_host(region_tag)}/check_token"

    params = {"locale": localize(locale)}
    data = {'token': token}

    return url, data, params
